@@ -55,12 +55,12 @@ def create_attack_graph(
     else:
         raise TypeError("`lang` must be either string or LanguageGraph")
 
-    if log_configs['langspec_file']:
+    if 'langspec_file' in log_configs:
         lang_graph.save_language_specification_to_json(
             log_configs['langspec_file']
         )
 
-    if log_configs['langgraph_file']:
+    if 'langgraph_file' in log_configs:
         lang_graph.save_to_file(log_configs['langgraph_file'])
 
     # Load model
@@ -89,7 +89,7 @@ def create_attack_graph(
 
 class AttackGraph():
     """Graph representation of attack steps"""
-    def __init__(self, lang_graph, model: Optional[Model] = None):
+    def __init__(self, lang_graph: LanguageGraph, model: Optional[Model] = None):
         self.nodes: dict[int, AttackGraphNode] = {}
         # Dictionaries used in optimization to get nodes by id or full name
         # faster
@@ -194,7 +194,7 @@ class AttackGraph():
                 ttc_dist = node_dict['ttc'],
                 existence_status = node_dict.get('existence_status', None)
             )
-            ag_node.tags = set(node_dict.get('tags', []))
+            ag_node.tags = list(node_dict.get('tags', []))
             ag_node.extras = node_dict.get('extras', {})
 
             if node_asset:
@@ -469,20 +469,24 @@ class AttackGraph():
                 existence_status = None
                 node_name = asset.name + ':' + attack_step.name
 
-                ttc_dist = attack_step.ttc
+                ttc_dist = copy.deepcopy(attack_step.ttc)
                 match (attack_step.type):
                     case 'defense':
                         # Set the TTC probability for defenses
-                        defense_value = float(asset.defenses[attack_step.name])
-                        ttc_dist = {
-                            'arguments': [defense_value],
-                            'name': 'Bernoulli',
-                            'type': 'function'
-                        }
-                        logger.debug(
-                            'Setting defense \"%s\" to "%s".',
-                            node_name, defense_value
-                        )
+                        # that were explicitly set in model
+                        if attack_step.name in asset.defenses:
+                            defense_value = float(
+                                asset.defenses[attack_step.name]
+                            )
+                            ttc_dist = {
+                                'arguments': [defense_value],
+                                'name': 'Bernoulli',
+                                'type': 'function'
+                            }
+                            logger.debug(
+                                'Setting defense \"%s\" to "%s".',
+                                node_name, defense_value
+                            )
 
                     case 'exist' | 'notExist':
                         # Resolve step expression associated with
@@ -531,15 +535,15 @@ class AttackGraph():
             if not ag_node.model_asset:
                 raise AttackGraphException('Attack graph node is missing '
                     'asset link')
-            lang_graph_asset = self.lang_graph.assets[
-                ag_node.model_asset.type]
 
-            lang_graph_attack_step = lang_graph_asset.attack_steps[
-                ag_node.name]
+            lang_graph_asset = self.lang_graph.assets[ag_node.model_asset.type]
+            lang_graph_attack_step: Optional[LanguageGraphAttackStep] = (
+                lang_graph_asset.attack_steps[ag_node.name]
+            )
 
             while lang_graph_attack_step:
-                for child in lang_graph_attack_step.children.values():
-                    for target_attack_step, expr_chain in child:
+                for target_attack_step, expr_chains in lang_graph_attack_step.children.items():
+                    for expr_chain in expr_chains:
                         target_assets = self._follow_expr_chain(
                             self.model,
                             set([ag_node.model_asset]),
