@@ -26,83 +26,6 @@ from ..exceptions import (
 
 logger = logging.getLogger(__name__)
 
-predef_ttcs: dict[str, dict] = {
-    'EasyAndUncertain':
-    {
-        'arguments': [0.5],
-        'name': 'Bernoulli',
-        'type': 'function'
-    },
-    'HardAndUncertain':
-    {
-        'lhs':
-        {
-            'arguments': [0.1],
-            'name': 'Exponential',
-            'type': 'function'
-        },
-        'rhs':
-        {
-            'arguments': [0.5],
-            'name': 'Bernoulli',
-            'type': 'function'
-        },
-        'type': 'multiplication'
-    },
-    'VeryHardAndUncertain':
-    {
-        'lhs':
-        {
-            'arguments': [0.01],
-            'name': 'Exponential',
-            'type': 'function'
-        },
-        'rhs':
-        {
-            'arguments': [0.5],
-            'name': 'Bernoulli',
-            'type': 'function'
-        },
-        'type': 'multiplication'
-    },
-    'EasyAndCertain':
-    {
-        'arguments': [1.0],
-        'name': 'Exponential',
-        'type': 'function'
-    },
-    'HardAndCertain':
-    {
-        'arguments': [0.1],
-        'name': 'Exponential',
-        'type': 'function'
-    },
-    'VeryHardAndCertain':
-    {
-        'arguments': [0.01],
-        'name': 'Exponential',
-        'type': 'function'
-    },
-    'Enabled':
-    {
-        'arguments': [1.0],
-        'name': 'Bernoulli',
-        'type': 'function'
-    },
-    'Instant':
-    {
-        'arguments': [1.0],
-        'name': 'Bernoulli',
-        'type': 'function'
-    },
-    'Disabled':
-    {
-        'arguments': [1.0],
-        'name': 'Bernoulli',
-        'type': 'function'
-    },
-}
-
 
 def disaggregate_attack_step_full_name(
         attack_step_full_name: str) -> list[str]:
@@ -435,20 +358,58 @@ class LanguageGraphAttackStep:
     name: str
     type: str
     asset: LanguageGraphAsset
-    ttc: dict = field(default_factory = dict)
+    ttc: Optional[dict] = field(default_factory = dict)
     overrides: bool = False
-    children: dict = field(default_factory = dict)
-    parents: dict = field(default_factory = dict)
+
+    own_children: dict[LanguageGraphAttackStep, list[ExpressionsChain | None]] = (
+        field(default_factory = dict)
+    )
+    own_parents: dict[LanguageGraphAttackStep, list[ExpressionsChain | None]] = (
+        field(default_factory = dict)
+    )
     info: dict = field(default_factory = dict)
     inherits: Optional[LanguageGraphAttackStep] = None
     own_requires: list[ExpressionsChain] = field(default_factory=list)
-    tags: set = field(default_factory = set)
+    tags: list = field(default_factory = list)
     detectors: dict = field(default_factory = lambda: {})
 
 
     def __hash__(self):
         return hash(self.full_name)
 
+    @property
+    def children(self) -> dict[
+        LanguageGraphAttackStep, list[ExpressionsChain | None]
+    ]:
+        """
+        Get all (both own and inherited) children of a LanguageGraphAttackStep
+        """
+
+        all_children = dict(self.own_children)
+
+        if self.overrides:
+            # Override overrides the children
+            return all_children
+
+        if not self.inherits:
+            return all_children
+
+        for child_step, chains in self.inherits.children.items():
+            if child_step in all_children:
+                all_children[child_step] += [
+                    chain for chain in chains
+                    if chain not in all_children[child_step]
+                ]
+            else:
+                all_children[child_step] = chains
+
+        return all_children
+
+    @property
+    def parents(self) -> None:
+        raise NotImplementedError(
+            "Can not fetch parents of a LanguageGraphAttackStep"
+        )
 
     @property
     def full_name(self) -> str:
@@ -466,8 +427,8 @@ class LanguageGraphAttackStep:
             'type': self.type,
             'asset': self.asset.name,
             'ttc': self.ttc,
-            'children': {},
-            'parents': {},
+            'own_children': {},
+            'own_parents': {},
             'info': self.info,
             'overrides': self.overrides,
             'inherits': self.inherits.full_name if self.inherits else None,
@@ -476,25 +437,22 @@ class LanguageGraphAttackStep:
             self.detectors.items()},
         }
 
-        for child in self.children:
-            node_dict['children'][child] = []
-            for (_, expr_chain) in self.children[child]:
-                if expr_chain:
-                    node_dict['children'][child].append(
-                        expr_chain.to_dict())
+        for child, expr_chains in self.own_children.items():
+            node_dict['own_children'][child.full_name] = []
+            for chain in expr_chains:
+                if chain:
+                    node_dict['own_children'][child.full_name].append(chain.to_dict())
                 else:
-                    node_dict['children'][child].append(None)
-
-        for parent in self.parents:
-            node_dict['parents'][parent] = []
-            for (_, expr_chain) in self.parents[parent]:
-                if expr_chain:
-                    node_dict['parents'][parent].append(
-                        expr_chain.to_dict())
+                    node_dict['own_children'][child.full_name].append(None)
+        for parent, expr_chains in self.own_children.items():
+            node_dict['own_parents'][parent.full_name] = []
+            for chain in expr_chains:
+                if chain:
+                    node_dict['own_parents'][parent.full_name].append(chain.to_dict())
                 else:
-                    node_dict['parents'][parent].append(None)
+                    node_dict['own_parents'][parent.full_name].append(None)
 
-        if hasattr(self, 'own_requires'):
+        if self.own_requires:
             node_dict['requires'] = []
             for requirement in self.own_requires:
                 node_dict['requires'].append(requirement.to_dict())
@@ -728,7 +686,6 @@ class LanguageGraph():
     """Graph representation of a MAL language"""
     def __init__(self, lang: Optional[dict] = None):
         self.assets: dict[str, LanguageGraphAsset] = {}
-        self.predef_ttcs: dict[str, dict] = predef_ttcs
         if lang is not None:
             self._lang_spec: dict = lang
             self.metadata = {
@@ -769,34 +726,6 @@ class LanguageGraph():
             langspec = archive.read('langspec.json')
             return LanguageGraph(json.loads(langspec))
 
-
-    def replace_if_predef_ttc(
-        self,
-        ttc_entry: Optional[dict]
-    ) -> dict:
-        """
-        If the TTC provided is a predefined name replace it with the
-        probability distribution it corresponds to. Otherwise, simply return
-        the TTC distribution provided as is.
-
-        Arguments:
-        ttc_entry   - the TTC entry to check for predefined names
-
-        Returns:
-        If the TTC entry provided contained a predefined name the TTC
-        probability distrubtion corresponding to it. Otherwise, the TTC
-        distribution provided as a parameter as is.
-        """
-        if ttc_entry is None:
-            return {}
-
-        ttc = self.predef_ttcs.get(str(ttc_entry.get('name')))
-        if ttc is not None:
-            return ttc
-        else:
-            return ttc_entry
-
-
     def _to_dict(self):
         """Converts LanguageGraph into a dict"""
 
@@ -940,18 +869,16 @@ class LanguageGraph():
                 asset_dict['name']
             )
             for attack_step_dict in asset_dict['attack_steps'].values():
-                ttc = lang_graph.replace_if_predef_ttc(
-                    attack_step_dict['ttc'])
                 attack_step_node = LanguageGraphAttackStep(
                     name = attack_step_dict['name'],
                     type = attack_step_dict['type'],
                     asset = asset,
-                    ttc = ttc,
+                    ttc = attack_step_dict['ttc'],
                     overrides = attack_step_dict['overrides'],
-                    children = {},
-                    parents = {},
+                    own_children = {},
+                    own_parents = {},
                     info = attack_step_dict['info'],
-                    tags = set(attack_step_dict['tags'])
+                    tags = list(attack_step_dict['tags'])
                 )
                 asset.attack_steps[attack_step_dict['name']] = \
                     attack_step_node
@@ -978,12 +905,11 @@ class LanguageGraph():
             asset = lang_graph.assets[asset_dict['name']]
             for attack_step_dict in asset_dict['attack_steps'].values():
                 attack_step = asset.attack_steps[attack_step_dict['name']]
-                for child_target in attack_step_dict['children'].items():
+                for child_target in attack_step_dict['own_children'].items():
                     target_full_attack_step_name = child_target[0]
                     expr_chains = child_target[1]
                     target_asset_name, target_attack_step_name = \
-                        disaggregate_attack_step_full_name(
-                            target_full_attack_step_name)
+                        disaggregate_attack_step_full_name(target_full_attack_step_name)
                     target_asset = lang_graph.assets[target_asset_name]
                     target_attack_step = target_asset.attack_steps[
                         target_attack_step_name]
@@ -992,33 +918,30 @@ class LanguageGraph():
                             expr_chain_dict,
                             lang_graph
                         )
-                        if target_attack_step.full_name in attack_step.children:
-                            attack_step.children[target_attack_step.full_name].\
-                                append((target_attack_step, expr_chain))
-                        else:
-                            attack_step.children[target_attack_step.full_name] = \
-                                [(target_attack_step, expr_chain)]
 
-                for parent_target in attack_step_dict['parents'].items():
-                    target_full_attack_step_name = parent_target[0]
-                    expr_chains = parent_target[1]
+                        if target_attack_step in attack_step.own_children:
+                            attack_step.own_children[target_attack_step].append(expr_chain)
+                        else:
+                            attack_step.own_children[target_attack_step] = [expr_chain]
+
+                for (target_step_full_name, expr_chains) in attack_step_dict['own_parents'].items():
                     target_asset_name, target_attack_step_name = \
                         disaggregate_attack_step_full_name(
-                            target_full_attack_step_name)
+                            target_step_full_name
+                        )
                     target_asset = lang_graph.assets[target_asset_name]
-                    target_attack_step = target_asset.attack_steps[
-                        target_attack_step_name]
+                    target_attack_step = target_asset.attack_steps[target_attack_step_name]
                     for expr_chain_dict in expr_chains:
                         expr_chain = ExpressionsChain._from_dict(
-                            expr_chain_dict,
-                            lang_graph
+                            expr_chain_dict, lang_graph
                         )
-                        if target_attack_step.full_name in attack_step.parents:
-                            attack_step.parents[target_attack_step.full_name].\
-                                append((target_attack_step, expr_chain))
+
+                        if target_attack_step in attack_step.own_parents:
+                            attack_step.own_parents[target_attack_step].append(
+                                expr_chain
+                            )
                         else:
-                            attack_step.parents[target_attack_step.full_name] = \
-                                [(target_attack_step, expr_chain)]
+                            attack_step.own_parents[target_attack_step] = [expr_chain]
 
                 # Recreate the requirements of exist and notExist attack steps
                 if attack_step.type == 'exist' or \
@@ -1608,6 +1531,7 @@ class LanguageGraph():
         """
         Link assets based on inheritance and associations.
         """
+
         for asset_dict in lang_spec['assets']:
             asset = assets[asset_dict['name']]
             if asset_dict['superAsset']:
@@ -1663,20 +1587,19 @@ class LanguageGraph():
                     attack_step_attribs['name']
                 )
 
-                ttc = self.replace_if_predef_ttc(attack_step_attribs['ttc'])
                 attack_step_node = LanguageGraphAttackStep(
                     name = attack_step_attribs['name'],
                     type = attack_step_attribs['type'],
                     asset = asset,
-                    ttc = ttc,
+                    ttc = attack_step_attribs['ttc'],
                     overrides = (
                         attack_step_attribs['reaches']['overrides']
                         if attack_step_attribs['reaches'] else False
                     ),
-                    children = {},
-                    parents = {},
+                    own_children = {},
+                    own_parents = {},
                     info = attack_step_attribs['meta'],
-                    tags = set(attack_step_attribs['tags'])
+                    tags = list(attack_step_attribs['tags'])
                 )
                 langspec_dict[attack_step_node.full_name] = \
                     attack_step_attribs
@@ -1716,10 +1639,10 @@ class LanguageGraph():
                                 asset = asset,
                                 ttc = attack_step.ttc,
                                 overrides = False,
-                                children = {},
-                                parents = {},
+                                own_children = {},
+                                own_parents = {},
                                 info = attack_step.info,
-                                tags = set(attack_step.tags)
+                                tags = list(attack_step.tags)
                             )
                             attack_step_node.inherits = attack_step
                             asset.attack_steps[attack_step.name] = attack_step_node
@@ -1727,12 +1650,9 @@ class LanguageGraph():
                             # The inherited attack step was already overridden.
                             continue
                         else:
-                            asset.attack_steps[attack_step.name].inherits = \
-                                attack_step
-                            asset.attack_steps[attack_step.name].tags |= \
-                                attack_step.tags
-                            asset.attack_steps[attack_step.name].info |= \
-                                attack_step.info
+                            asset.attack_steps[attack_step.name].inherits = attack_step
+                            asset.attack_steps[attack_step.name].tags += attack_step.tags
+                            asset.attack_steps[attack_step.name].info |= attack_step.info
 
         # Then, link all of the attack step nodes according to their
         # associations.
@@ -1786,16 +1706,14 @@ class LanguageGraph():
                         target_attack_step_name]
 
                     # Link to the children target attack steps
-                    attack_step.children.setdefault(target_attack_step.full_name, [])
-                    attack_step.children[target_attack_step.full_name].append(
-                        (target_attack_step, expr_chain)
-                    )
+                    attack_step.own_children.setdefault(target_attack_step, [])
+                    attack_step.own_children[target_attack_step].append(expr_chain)
 
                     # Reverse the children associations chains to get the
                     # parents associations chain.
-                    target_attack_step.parents.setdefault(attack_step.full_name, [])
-                    target_attack_step.parents[attack_step.full_name].append(
-                        (attack_step, self.reverse_expr_chain(expr_chain, None))
+                    target_attack_step.own_parents.setdefault(attack_step, [])
+                    target_attack_step.own_parents[attack_step].append(
+                        self.reverse_expr_chain(expr_chain, None)
                     )
 
                 # Evaluate the requirements of exist and notExist attack steps
